import csv
import json
import jinja2 as jj
from pathlib import Path

from .markdown_processing import process_markdown_text
from ..util import parse_soup_from_xml, retrieve_contents


def _read_md_table(md_text: str) -> list[dict]:
    # Assume the only thing in the markdown is a single table
    html = process_markdown_text(md_text)

    soup = parse_soup_from_xml(html)
    table = soup.find('table')

    # Thank you, GPT:
    # Extract headers
    headers = [th.text.strip() for th in table.find_all('th')]

    # Iterate over rows and create list of dictionaries
    rows = []
    for tr in table.find_all('tr')[1:]:  # skipping the header row
        cells = tr.find_all(['td', 'th'])
        if len(cells) != len(headers):
            continue  # Skip any rows that do not match the number of headers
        row_data = {headers[i]: retrieve_contents(cells[i]) for i in range(len(headers))}
        rows.append(row_data)

    return rows


def _get_global_args(global_args_path: Path) -> dict:
    if '.json' in global_args_path.name:
        return json.loads(global_args_path.read_text())
    elif '.csv' in global_args_path.name:
        return dict(csv.DictReader(global_args_path.read_text().splitlines()))
    else:
        raise NotImplementedError('Global args file of type: ' + global_args_path.suffix)


def _get_args(args_path: Path) -> list[dict]:
    if args_path.suffix == '.json':
        return json.loads(args_path.read_text())

    elif args_path.suffix == '.csv':
        return list(csv.DictReader(args_path.read_text().splitlines()))

    elif args_path.suffix == '.md':
        return _read_md_table(args_path.read_text())

    else:
        raise NotImplementedError('Args file of type: ' + args_path.suffix)


def _render_template(template, **kwargs):
    jj_template = jj.Environment().from_string(template)
    kwargs |= dict(zip=zip, split_list=lambda x: x.split(';'))
    return jj_template.render(**kwargs)


def _process_template(template: str, arg_sets: list[dict]):
    return '\n'.join([_render_template(template, **args) for args in arg_sets])


def process_jinja(
        template: str,
        args_path: Path = None,
        global_args_path: Path = None,
        **kwargs
) -> str:
    arg_sets = _get_args(args_path) if args_path is not None else None

    if global_args_path:
        kwargs |= _get_global_args(global_args_path)

    if arg_sets is not None:
        arg_sets = [{**args, **kwargs} for args in arg_sets]
    else:
        arg_sets = [kwargs]

    return _process_template(template, arg_sets)



import logging

logger = None

def get_logger(name: str = None):
    global logger
    if logger is None:
        logger = logging.getLogger('MDXCANVAS')
        logger.setLevel(logging.INFO)

        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)

        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        ch.setFormatter(formatter)

        logger.addHandler(ch)
    if name:
        for handler in logger.handlers:
            if isinstance(handler, logging.StreamHandler):
                formatter = logging.Formatter(f'%(asctime)s - {name} - %(name)s - %(levelname)s - %(message)s')
                handler.setFormatter(formatter)
                break

    return logger



import logging
import re
from pathlib import Path
from tempfile import TemporaryDirectory
from zipfile import ZipFile, ZipInfo

from canvasapi.course import Course

from ..resources import ZipFileData, FileData
from .file import deploy_file, lookup_file


def zip_folder(
        folder_path: Path,
        path_to_zip: Path,
        exclude: re.Pattern = None,
        priority_fld: Path = None
):
    """
    Zips a folder, excluding files that match the exclude pattern.
    Items from the priority folder are added to the zip if they are not in the standard folder.
    Items in the priority folder take precedence over items in the standard folder.
    """
    with ZipFile(path_to_zip, "w") as zipf:
        for item in folder_path.glob("*"):
            write_item_to_zip(item, zipf, exclude, priority_fld=priority_fld)


def write_item_to_zip(item: Path, zipf: ZipFile, exclude: re.Pattern = None, prefix='', priority_fld: Path = None):
    if exclude and exclude.match(item.name):
        logging.debug(f"Excluding file {item.name}")
        return
    if item.is_dir():
        write_directory(item, zipf, exclude, prefix, priority_fld / item.name)
    else:
        write_file(item, zipf, prefix, priority_fld)


def write_directory(folder: Path, zipf: ZipFile, exclude: re.Pattern = None, prefix='',
                    priority_fld: Path = None):
    prefix = prefix + folder.name + '/'

    # Get all items in the folder
    paths = list(folder.glob("*"))

    # Add items from priority folder that are not in the folder
    item_names = {i.name for i in folder.glob("*")}
    if priority_fld:
        for item in priority_fld.glob("*"):
            if item.name not in item_names:
                paths.append(item)
                logging.debug(f"Using additional file {item.name}")

    for path in paths:
        write_item_to_zip(path, zipf, exclude, prefix, priority_fld)


def set_time_1980(file, prefix=''):
    """
    Ensures that the zip file stays consistent between runs.
    """
    zinfo = ZipInfo(
        prefix + file.name,
        date_time=(1980, 1, 1, 0, 0, 0)
    )
    return zinfo


def write_file(file: Path, zipf: ZipFile, prefix='', priority_fld: Path = None):
    # Use the file from the priority folder if it exists
    if priority_fld and (priority_file := priority_fld / file.name).exists():
        file = priority_file
        logging.debug(f"Prioritizing file {file.name}")

    # For consistency, set the time to 1980
    zinfo = set_time_1980(file, prefix)
    try:
        with open(file) as f:
            zipf.writestr(zinfo, f.read())
    except UnicodeDecodeError as _:
        with open(file, 'rb') as f:
            zipf.writestr(zinfo, f.read())


def predeploy_zip(zipdata: ZipFileData, tmpdir: Path) -> FileData:
    target_folder = Path(zipdata['content_folder'])
    pf = zipdata['priority_folder']
    priority_folder = Path(pf) if pf is not None else None
    if priority_folder is not None and not priority_folder.exists():
        raise FileNotFoundError(priority_folder)

    exclude = re.compile(zipdata['exclude_pattern']) if zipdata['exclude_pattern'] is not None else None

    path_to_zip = tmpdir / zipdata['zip_file_name']
    zip_folder(target_folder, path_to_zip, exclude, priority_folder)

    file = FileData(
        path=str(path_to_zip),
        canvas_folder=zipdata['canvas_folder']
    )

    return file


deploy_zip = deploy_file
lookup_zip = lookup_file
